import math
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, LogLocator
from matplotlib.transforms import ScaledTranslation
import pandas as pd
import dask
import dask.dataframe as dd
import seaborn as sns

from atlasify import atlasify

from nnfwtbn.stack import Stack
from nnfwtbn.process import Process
from nnfwtbn.cut import Cut
from nnfwtbn.variable import Variable, BlindingStrategy
import nnfwtbn.error as err

# Store built-in range
range_ = range

ATLAS = "Internal"
INFO = "$\sqrt{s} = 13\,\mathrm{TeV}$, $139\,\mathrm{fb}^{-1}$\nnnfwtbn"

def fill_labels(label, info):
    if label is None:
        label = ATLAS
    if info is None:
        info = INFO
    return label, info

class HistogramFactory:
    """
    Short-cut to create multiple histogram with the same set of processes or
    in the same region.
    """

    def __init__(self, *args, **kwds):
        """
        Accepts any number of positional and keyword arguments. The arguments
        are stored internally and use default value for hist(). See __call__().
        """
        self.args = args
        self.kwds = kwds

    def __call__(self, *args, **kwds):
        """
        Proxy for method to hist(). The positional argument passed to hist()
        are the positional argument given to the constructor concatenated with
        the positional argument given to this method. The keyword argument for
        hist() is the union of the keyword arguments passed to the constructor
        and this method. The argument passed to this method have precedence.

        The method returns the return value of hist.
        """
        return hist(*self.args, *args, **dict(self.kwds, **kwds))

def hist(dataframe, variable, bins, stacks, selection=None,
         range=None, blind=None, axes=None, figure=None,
         weight=None, y_log=False, y_min=None, vlines=[],
         denominator=0, numerator=-1, ratio_label=None, diff=False,
         ratio_range=None, atlas=None, info=None, enlarge=1.6,
         density=False, include_outside=False, **kwds):
    """
    Creates a histogram of stacked processes. The first argument is the
    dataframe to operate on. The 'variable' argument defines the x-axis. The
    variable argument can be a Variable object or a string naming a column in
    the dataframe.

    The 'bins' argument can be an integer specifying the number of bins or a
    list with all bin boundaries. If it is an integer, the argument range is
    mandatory. The range argument must be a tuple with the lowest and highest
    bin edge. The properties of a Variable object are used for the x- and
    y-axis labels.

    Stacks must be Stack objects. The plotting style is defined via the stack
    object.

    The optional blind argument controls which stack should be blinded. The
    argument can be a single stack, a list of stacks or None. By default,
    no stack is blinded.

    If the figure argument is omitted, this method creates a new axes and
    figure. If axes only is omitted, the method creates a new axes from the
    figure. If a ratio plot is drawn, axes must a list if main, ratio plot if
    not omitted.

    The method returns (figure, axes) which were used during plotting. This
    might be identical to the figure and axes arguments. If a ratio plot is
    drawn, the axes return value is a list of main, ratio plot.

    The weight is used to weight the entries. Entries have unit
    weight if omitted. The argument can be a string name of a column or a
    variable object.

    If the y_log argument is set to True, the y axis will be logarithmic. The
    axis is enlarged on a logarithmic scale to make room for the ATLAS labels.
    The optional y_min argument can be used to set the lower limit of the y
    axis. The default is 0 for linear scale, and 1 for logarithmic scale.

    The option vlines can be used to draw vertical lines onto the histogram,
    e.g., a cut line. The argument should be an array, one item for each line.
    If the item is a number a red line will be drawn at that x-position. If it
    is a dict, it will take the item 'x' to determine the position, all other
    keywords are passed to matplotlibs plot method.`

    The numerator and denominator arguments control the content of the ratio
    plot. The arguments can be either an integer or a list of processes. If
    an argument is an integer, it is used as an index on the stack list. If
    an argument is a custom stack, it will be evaluated and
    used for the ratio plot. If either is None, no ratio plot will be drawn.
    The default draws the ratio between the first and the last stack. If a
    ratio plot is drawn, the axes argument must be None or a list of main, ratio
    axes.

    The ratio_label option controls the label of the ratio plot.

    The ratio_range argument control the y-range of the ratio plot. If set to
    None, it will scale automatically to include all points. The default is
    is None.

    If diff is set to True, The difference between the 'numerator' and the
    'denominator' is down instead of their ratio.

    The module constants ATLAS and INFO are passed to atlasify. Overwrite them
    to change the badges.

    If the density argument is True, the area of each stack is normalized to
    unity.
    """
    # Wrap column string by variable
    if isinstance(variable, str):
        variable = Variable(variable, variable)

    if weight is None:
        weight = Variable("unity", lambda d: variable(d) * 0 + 1)
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    squared_weight = Variable("squared weight", lambda d: weight(d)**2)

    draw_ratio = (denominator is not None) and (numerator is not None)

    # Disable ratio plots if there is only one stack 
    if len(stacks) == 1 and \
            isinstance(denominator, int) and \
            isinstance(numerator, int):
        draw_ratio = False

    # Handle axes, figure
    if figure is None or axes is None:
        if figure is None:
            provider = plt
        else:
            provider = figure

        if draw_ratio:
            # Make plot area (without enlargement) match golden ratio
            subplot_kwds = {"gridspec_kw": {"height_ratios": [3, 1]},
                            "figsize": (5.75, 5.75),
                            "sharex": True}
            figure, (axes, axes_ratio) = provider.subplots(2, 1,
                                                        **subplot_kwds)

        else:
            figure, axes = provider.subplots(figsize=(5.75, 5.75 / 4 * 3))
    all_axes = [axes]
    if draw_ratio:
        all_axes.append(axes_ratio)

    # Handle selection
    if selection is None:
        selection = Cut(lambda d: variable(d) * 0 == 0)
    elif not isinstance(selection, Cut):
        selection = Cut(selection)

    # Handle range/bins
    equidistant_bins = False
    if range is not None:
        # Build bins
        if not isinstance(bins, int):
            raise err.InvalidBins("When range is given, bins must be int.")
        if not isinstance(range, tuple)  or len(range) != 2:
            raise err.InvalidProcessSelection("Range argument must be a "
                                              "tuple of two numbers.")
        bins = np.linspace(range[0], range[1], bins + 1)
        equidistant_bins = True

    bin_centers = (bins[1:] + bins[:-1]) / 2
    bin_widths = bins[1:] - bins[:-1]

    # # Check structure of kwds
    # new_kwds = {}
    # for kwd, value in kwds.items():
    #     if isinstance(value, list):
    #         if len(value) != len(stacks):
    #             raise ValueError("Length of %s must equal number of stacks."
    #                              % repr(kwd))

    #         # Wrap properties for the whole stack in a list
    #         new_kwds[kwd] = [(_ if isinstance(_, list) else [_]) for _ in value] 

    #     else:
    #         # Single value for all stacks and all processes
    #         new_kwds[kwd] = [[value]]*len(stacks)
    # kwds = new_kwds

    numerator_hist = None
    denominator_hist = None

    if blind is None:
        blind = []
    elif isinstance(blind, Stack):
        # Wrap scalar blind stack
        blind = [blind]

    histograms = []
    uncertainties = []
    process_index_orders = []

    # Handle stack
    max_content = float('-inf')
    for i_stack, stack in enumerate(stacks):
        uncertainty = np.array(0)
        if stack in blind and variable.blinding is not None:
            c_blind = variable.blinding(variable, bins)
        else:
            c_blind = lambda d: d

        density_norm = 1.0
        if density:
            density_norm = stack.get_total(dataframe,
                                           [float('-inf'), float('inf')],
                                           variable, weight, include_outside).sum()
             
        if y_log:
            total_yields = []
            for i_process in range_(len(stack)):
                total_yield = stack.get_hist(dataframe, i_process,
                                             [float('-inf'), float('inf')],
                                             variable, weight,
                                             include_outside).sum()
                total_yields.append((i_process, total_yield))

            total_yields = list(dask.compute(*total_yields))
            total_yields.sort(key=lambda x: x[1])
            process_index_order = [x[0] for x in total_yields]
        else:
            process_index_order = range_(len(stack))

        process_index_orders.append(process_index_order)
        for i_process in process_index_order:
            process = stack.processes[i_process]
            histogram = stack.get_hist(c_blind(dataframe), i_process, bins,
                                       variable, weight, include_outside)

            histogram = histogram / density_norm

            # Draw uncertainty iff this is the last processes of the stack
            if i_process == process_index_order[-1]:
                uncertainty = stack.get_total_uncertainty(c_blind(dataframe),
                                                          bins, variable,
                                                          weight,
                                                          include_outside)
                uncertainty = uncertainty / density_norm

            histograms.append(histogram)
            uncertainties.append(uncertainty)


    # Compute delayed dask 
    histograms, uncertainties = dask.compute(histograms, uncertainties)

    # Reverse to use pop()
    histograms.reverse()
    uncertainties.reverse()
    process_index_orders.reverse()

    for i_stack, stack in enumerate(stacks):
        bottom = 0
        process_index_order = process_index_orders.pop()
        for i_process in process_index_order:
            process = stack.processes[i_process]

            histogram = histograms.pop()
            uncertainty = uncertainties.pop()

            # Draw process
            histtype = stack.get_histtype(i_process)
            if histtype == "points":
                kwds = {'markersize': 4, 'fmt': 'o'}
                kwds.update(stack.get_aux(i_process))

                non_empty = (histogram != 0)

                axes.errorbar(bin_centers[non_empty],
                              (bottom + histogram)[non_empty],
                              uncertainty[non_empty],
                              (bin_widths / 2)[non_empty],
                              label=process.label, **kwds)
            else:
                kwds = {}
                kwds.update(stack.get_aux(i_process))
                axes.hist(bin_centers, bins=bins, bottom=bottom,
                          label=process.label, weights=histogram,
                          histtype=histtype, **kwds)

                # Draw uncertainty band
                if (uncertainty > 0).any():
                    band_lower = bottom + histogram - uncertainty
                    axes.hist(bin_centers, bins=bins, bottom=band_lower,
                              weights=uncertainty * 2, fill=False, hatch='/////',
                              linewidth=0, edgecolor="#666666", label="Stat.  uncertainty")

            bottom += histogram
        # (end of process loop)

        # Track highest point
        max_content = max(max_content, np.max(bottom + uncertainty / 2))

        # Resolve numerator/denominator indices
        if isinstance(numerator, int) and stacks[numerator] == stack:
            numerator = stack
        if isinstance(denominator, int) and stacks[denominator] == stack:
            denominator= stack
    # (end of stack loop)

    # Handle ratio plot
    if draw_ratio:
        _draw_ratio(dataframe, numerator, denominator, bins, variable, weight,
                    axes_ratio, diff, ratio_range, ratio_label, blind,
                    include_outside)

    # Draw vertical lines
    for vline in vlines:
        try: 
            x = vline['x']
            kwds = {k: v for k, v in vline.items() if k != 'x'}
        except TypeError:
            x = vline
            kwds = {'color': 'r'}
        for ax in all_axes:
            ax.plot([x, x], ax.get_ylim(), **kwds)

    if (atlas is not False) or (info is not False):
        atlasify(*fill_labels(atlas, info), enlarge=1.0, axes=axes)

    # Configure x-axis
    axes.set_xlim((bins.min(), bins.max()))

    # Configure y-axis
    _, y_max = axes.get_ylim()

    if y_log:
        y_min = y_min if y_min is not None else 1
        y_max = max([y_max, 10**enlarge * max_content])
        axes.set_yscale('log')
    else:
        y_min = y_min if y_min is not None else 0
        y_max = max([y_max, enlarge * max_content])

    axes.set_ylim((y_min, y_max))

    leg_handles, leg_labels = axes.get_legend_handles_labels()
    axes.legend(leg_handles[::-1], leg_labels[::-1], frameon=False, loc=1)

    x_label_axes = axes_ratio if draw_ratio else axes
    if variable.unit is not None:
        x_label_axes.set_xlabel("%s / %s" % (variable.name, variable.unit),
                        horizontalalignment='right', x=0.95)
    else:
        x_label_axes.set_xlabel(variable.name,
                                horizontalalignment='right', x=0.95)

    if equidistant_bins:
        subject = "Fraction" if density else "Events"
        if variable.unit is not None:
            axes.set_ylabel(f"{subject} / %g %s" % (bins[1] - bins[0], variable.unit),
                            horizontalalignment='right', y=0.95)
        else:
            axes.set_ylabel(f"{subject} / %g" % (bins[1] - bins[0]),
                            horizontalalignment='right', y=0.95)
    else:
        axes.set_ylabel("Events / Bin", horizontalalignment='right', y=0.95)

    for ax in all_axes:
        ax.tick_params("both", which="both", direction="in")
        ax.tick_params("both", which="major", length=6)
        ax.tick_params("both", which="minor", length=3)
        ax.tick_params("x", which="both", top=True)
        ax.tick_params("y", which="both", right=True)
        ax.xaxis.set_minor_locator(AutoMinorLocator())

    if not y_log:
        axes.yaxis.set_minor_locator(AutoMinorLocator())


    if draw_ratio:
        figure.subplots_adjust(hspace=0.025)
        return figure, (axes, axes_ratio)
    else:
        return figure, axes

def _draw_ratio(df, numerator, denominator, bins, variable, weight,
                axes_ratio, diff=False, ratio_range=None, ratio_label=None,
                blind=[], include_outside=False):

    if numerator in blind and variable.blinding is not None:
        c_blind = variable.blinding(variable, bins)
    else:
        c_blind = lambda d: d
    numerator_hist = numerator.get_total(c_blind(df), bins, variable, weight,
                                         include_outside)
    numerator_error = numerator.get_total_uncertainty(c_blind(df), bins,
                                                      variable, weight,
                                                      include_outside)

    if denominator in blind and variable.blinding is not None:
        c_blind = variable.blinding(variable, bins)
    else:
        c_blind = lambda d: d
    denominator_hist = denominator.get_total(c_blind(df), bins, variable,
                                             weight, include_outside)
    denominator_error = denominator.get_total_uncertainty(c_blind(df), bins,
                                                          variable, weight,
                                                          include_outside)

    data = numerator_hist, numerator_error, denominator_hist, denominator_error
    data = dask.compute(*data)
    numerator_hist, numerator_error, denominator_hist, denominator_error = data
         

    # valid bins with non-zero denominator
    non_empty = (denominator_hist != 0) * (numerator_hist != 0)
    if diff:
        non_empty = (numerator_hist != 0)

    bin_centers = ((bins[1:] + bins[:-1]) / 2)[non_empty]
    bin_widths = (bins[1:] - bins[:-1])[non_empty]

    if diff:
        ratio = numerator_hist[non_empty] - denominator_hist[non_empty]
        ratio_error = numerator_error[non_empty]
    else:
        ratio = numerator_hist[non_empty] / denominator_hist[non_empty]
        ratio_error = numerator_error[non_empty] / denominator_hist[non_empty]

    axes_ratio.errorbar(bin_centers,
                        ratio,
                        ratio_error,
                        bin_widths / 2,
                        color='k',
                        markersize=4, fmt='o')

    if diff:
        non_empty = np.ones(len(numerator_hist), dtype='bool')
        rel_error = denominator_error
        band_lower = 0 - rel_error
    else:
        non_empty = (denominator_hist != 0)
        denominator_hist[~non_empty] = 1
        numerator_error[non_empty] = 0
        rel_error = denominator_error / denominator_hist
        band_lower = 1 - rel_error

    axes_ratio.hist(bins[:-1][non_empty],
                    bins=bins,
                    bottom=band_lower,
                    weights=(rel_error * 2)[non_empty],
                    fill=False,
                    hatch='/////',
                    linewidth=0,
                    edgecolor="#666666")

    if ratio_range is not None:
        axes_ratio.set_ylim(ratio_range)

    if ratio_label is not None:
        axes_ratio.set_ylabel(ratio_label)

def _transpose(array):
    """Return a transposed version of the array"""
    if array == [] or array == [[]]:
        return [[]]

    return list(map(list, zip(*array)))

def _dask_compute(array):
    """Call dask.compute() on flattened list and return the structured result"""
    if not array or not array[0]:
        return array

    x_len = len(array)
    y_len = len(array[0])

    flat = [i for s in array for i in s]
    flat = dask.compute(*flat)

    return [[flat[x * y_len + y] for y in range(y_len)] for x in range(x_len)]

def confusion_matrix(df, x_processes, y_processes, x_label, y_label,
                     weight=None, axes=None, figure=None, atlas=ATLAS,
                     info=None, enlarge=1.3, normalize_rows=False, **kwds):
    """
    Creates a confusion matrix.
    """
    # Wrap column string by variable
    if weight is None:
        weight = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    # Handle axes, figure
    if figure is None:
        figure, axes = plt.subplots()  #figsize=(5,5))
    elif axes is None:
        axes = figure.subplots()

    y_processes = y_processes[::-1]

    if normalize_rows:
        # Swap processes
        y_processes, x_processes = x_processes, y_processes

    # Need a nested-list (not numpy array) to store Dask scalars
    data = [[0 for p in x_processes] for q in y_processes]
    for i_x, x_process in enumerate(x_processes):
        x_df = x_process(df)
        total_weight = weight(x_df).sum()
        for i_y, y_process in enumerate(y_processes):
            x_y_df = y_process(x_df)
            data[i_y][i_x] = weight(x_y_df).sum() / total_weight

    if normalize_rows:
        # Swap processes back
        y_processes, x_processes = x_processes, y_processes
        data = _transpose(data)

    data = _dask_compute(data)

    data = pd.DataFrame(data,
                        columns=[p.label for p in x_processes],
                        index=[p.label for p in y_processes])
    if normalize_rows:
        # Swap labels
        y_label, x_label = x_label, y_label

    sns.heatmap(data, **dict(vmin=0, vmax=1, cmap="Greens", ax=axes,
                           cbar_kws={
                               'label': "$P($%s$|$%s$)$" % (y_label, x_label)
                               }
                           ), **kwds)
    if normalize_rows:
        # Swap labels back
        y_label, x_label = x_label, y_label
    axes.set_xlabel(x_label)
    axes.set_ylabel(y_label)
    axes.set_ylim(len(y_processes), 0)

    if (atlas is not False) or (info is not False):
        atlasify(*fill_labels(atlas, info), axes=axes, outside=True)
        figure.subplots_adjust(top=1/enlarge)

    return data


def roc(df, signal_process, background_process, discriminant, steps=100,
        selection=None, min=None, max=None, axes=None, weight=None,
        atlas=ATLAS, info=None, enlarge=1.3,
        return_auc=False):
    """
    Creates a ROC.

    The method returns a dataframe with the signal efficiency and background
    rejection columns. The length of the dataframe equals the steps parameter.

    If return_auc is True, the method returns a tuple with the area under the
    curve and an uncertainty estimation on the area.
    """
    # Wrap column string by variable
    if discriminant is None:
        discriminant = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(discriminant, str):
        discriminant = Variable(discriminant, discriminant)

    if weight is None:
        weight = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    # Handle selection
    if selection is None:
        selection = lambda df: df  # Identity
    elif not isinstance(selection, Cut):
        selection = Cut(selection)

    # Handle axes
    if axes is None:
        fig, axes = plt.subplots()
    if min is None:
        min = discriminant(df).min()
    if max is None:
        max = discriminant(df).max()

    df = selection(df)

    signal_effs = []
    background_ieffs = []
    n_total_sig = weight(signal_process.selection(df)).sum()
    n_total_bkg = weight(background_process.selection(df)).sum()
    for cut_value in np.linspace(min, max, steps):
        residual_df = df[discriminant(df) >= cut_value]

        n_total = weight(residual_df).sum()
        if n_total == 0:
            continue

        signal_df = signal_process.selection(residual_df)
        background_df = background_process.selection(residual_df)

        n_signal = weight(signal_df).sum()
        n_background = weight(background_df).sum()

        signal_effs.append(n_signal / n_total_sig)
        background_ieffs.append(1 - n_background / n_total_bkg)

    data = pd.DataFrame({"Signal efficiency": signal_effs,
                         "1 - Background efficiency": background_ieffs})
    sns.lineplot(x="Signal efficiency", y="1 - Background efficiency", data=data,
                 ax=axes, ci=None, label=discriminant.name)
    axes.plot([0, 1], [1, 0], color='gray', linestyle=':')

    axes.set_xlim((0, 1))
    axes.set_ylim((0, enlarge))
    axes.legend(loc=1, frameon=False)

    axes.tick_params("both", which="both", direction="in")
    axes.tick_params("both", which="major", length=6)
    axes.tick_params("both", which="minor", length=3)
    axes.tick_params("x", which="both", top=True)
    axes.tick_params("y", which="both", right=True)
    axes.xaxis.set_minor_locator(AutoMinorLocator())
    axes.yaxis.set_minor_locator(AutoMinorLocator())

    if (atlas is not False) or (info is not False):
        atlasify(*fill_labels(atlas, info), enlarge=1.0, axes=axes)

    if return_auc:
        # calculate rectangular approximations of AUC
        # one of these will be too high, the other too low
        # the difference gives a measure of uncertainty
        auc_approx_a = 0
        auc_approx_b = 0
        for i_bin in range(1,len(signal_effs)):
            bin_width = abs(signal_effs[i_bin] - signal_effs[i_bin-1])
            auc_approx_a += bin_width * background_ieffs[i_bin]
            auc_approx_b += bin_width * background_ieffs[i_bin - 1]
        auc_mean = (auc_approx_a + auc_approx_b)/2
        auc_diff = abs(auc_approx_a - auc_approx_b)/2
        auc_mean = auc_mean if auc_mean > 0.5 else 1 - auc_mean
        return (auc_mean, auc_diff)
    else:
        return data

def correlation_matrix(df, variables,
                       weight=None, axes=None, figure=None, atlas=ATLAS,
                       info=None, enlarge=1.3, normalize_rows=False, **kwds):
    """
    Plot the Pearson correlation coefficient matrix. The square matrix is
    returned as a DataFrame.
    """
    if weight is None:
        weight = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    kwds.setdefault("annot", True)
    kwds.setdefault("fmt", "2.0f")

    # Handle axes, figure
    if figure is None:
        figure, axes = plt.subplots()  #figsize=(5,5))
    elif axes is None:
        axes = figure.subplots()

    
    var_data = [v(df) for v in variables]
    var_data = dask.compute(*var_data)
    
    data = np.cov(np.array([v.to_numpy() for v in var_data]),
                  aweights=weight(df))
    var = np.diag(data)
    data = data / np.sqrt(var)
    data = data.T / np.sqrt(var)

    data = pd.DataFrame(data,
                        columns=[v.name for v in variables],
                        index=[v.name for v in variables])

    sns.heatmap(data * 100, **dict(vmin=-100, vmax=100, cmap="RdBu_r", ax=axes,
                           cbar_kws={
                               'label': "Correlation coefficient $\\rho \\cdot 100$"
                               }
                           ), **kwds)

    axes.set_ylim(len(variables), 0)

    if (atlas is not False) or (info is not False):
        atlasify(*fill_labels(atlas, info), axes=axes, outside=True)
        figure.subplots_adjust(top=1/enlarge)

    return data
